from pathlib import Path
import json
import os
import numpy as np
import pandas as pd


from .dtype import *


class DataHub(dict):
    """This class contains DataHub results."""

    def __init__(self, *args, **kwargs):
        if 'variable_type' in kwargs:
            
            super().__init__(**kwargs)
            self.variable_type = kwargs["variable_type"]
        else:
            raise ValueError("Unsupported initialization, missing visula type")


       


import random
import sys

from jinja2 import Environment, PackageLoader
from jinja2 import Template
from IPython.display import HTML

from .configs import *



class WizRenderer:
    def __init__(self, data_load, variable_type, cfg):
        #print(to_render)
        """
        {'type': ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm'], 
        'title': [None, None, None, None], 
        'image': [None, None, None, None], 
        'value_table': [None, None, None, None]}
        """
        self.cfg = cfg
        self.context = DataManager(**data_load)
        self.template = Template("""
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Matplotlib Plots</title>
    <style>
        /* Reset CSS -> Use a Reset CSS: to ensure consistent default styles across different browsers. */
        * {
            box-sizing: border-box;
            margin: 0;
            padding: 0;
        }

        /* Shared styles for tabset */
        .tabset-{{ context.unique_id }} {
            max-width: 1118px;
            padding-top: 15px;
        }

        .tabset-{{ context.unique_id }} > input[type="radio"] {
            position: absolute;
            left: -200vw;
        }

        .tabset-{{ context.unique_id }} .tab-panel {
            display: none;
            padding-top: 10px;
            border-top: 1px solid #ccc;
            width: 1100px;
        }

        .tabset-{{ context.unique_id }} > input:checked ~ .tab-panels > .tab-panel {
            display: block;
        }

        .tabset-{{ context.unique_id }} > label {
            position: relative;
            display: inline-block;
            padding: 15px 15px 25px;
            border: 1px solid transparent;
            border-bottom: 0;
            cursor: pointer;
            font-weight: 600;
            height: 64px;
            color: #525252;
        }

        input:focus-visible + label {
            outline: 2px solid rgba(0, 102, 204, 1);
            border-radius: 3px;
        }

        .tabset-{{ context.unique_id }} > label:hover,
        .tabset-{{ context.unique_id }} > input:focus + label,
        .tabset-{{ context.unique_id }} > input:checked + label {
           color: {{ cfg.label_color }};
        }

        .tabset-{{ context.unique_id }} > label:hover::after,
        .tabset-{{ context.unique_id }} > input:focus + label::after,
        .tabset-{{ context.unique_id }} > input:checked + label::after {
            background: #06c;
        }

        .tabset-{{ context.unique_id }} > input:checked + label {
            border-color: #ccc;
            border-bottom: 1px solid #fff;
            margin-bottom: -1px;
            height: 65px;
        }

        /* Unique body styling */
        .body-{{ context.unique_id }} {
            font: 16px/1.5em "Overpass", "Open Sans", Helvetica, sans-serif;
            color: #333;
            font-weight: 300;
            padding: 30px;
        }

    </style>
</head>
<body class="body-{{ context.unique_id }}">
    <div class="tabset-{{ context.unique_id }}">
        {% for idx in range(context.type|length) %}
            {% if context.title[idx] %}
                <input type="radio" name="tabset-{{ context.unique_id }}" id="{{ context.type[idx] }}-{{ context.unique_id }}" aria-controls="{{ context.type[idx] }}-{{ context.unique_id }}" {% if loop.first %}checked{% endif %}>
                <label for="{{ context.type[idx] }}-{{ context.unique_id }}">{{ context.title[idx] }}</label>
            {% endif %}
        {% endfor %}

        <div class="tab-panels">
            {% for idx in range(context.type|length) %}
                <section id="{{ context.type[idx] }}-{{ context.unique_id }}" class="tab-panel" {% if not loop.first %}style="display: none;"{% endif %}>
                    {% if context.value_table[idx] is not none %}
                        {{ context.value_table[idx]|safe }}
                    {% elif context.image[idx] is not none %}
                        {% for image_data in context.image[idx] %}
                            {% set image_width = image_data[1] | default(500) %}
                            <img src="data:image/png;base64, {{ image_data[0] }}" alt="{{ context.title[idx] }}" style="width: {{ image_width }}px; margin-right: 0px;">
                        {% endfor %}
                    {% endif %}
                </section>
            {% endfor %}
        </div>

        <script>
            {% for idx in range(context.type|length) %}
                document.getElementById("{{ context.type[idx] }}-{{ context.unique_id }}").addEventListener("change", function() {
                    showPanel("{{ context.type[idx] }}-{{ context.unique_id }}");
                });
            {% endfor %}

            function showPanel(panelId) {
                const panel = document.getElementById(panelId);
                const panels = panel.parentElement.querySelectorAll('.tab-panel');
                panels.forEach(p => {
                    p.style.display = p.id === panelId ? "block" : "none";
                });
            }
        </script>
    </div>
</body>
</html>
        """)


    def _repr_html_(self):
        output_html = self.template.render(context=self.context, cfg=self.cfg)
        return output_html



    def show(self):
        """
        Render the report. This is useful when calling plot in a for loop.
        """
        display(HTML(self._repr_html_()))

    
class DataManager:
    def __init__(self, **data_load):
        for key, value in data_load.items():
            setattr(self, key, value)
        self.unique_id = random.randint(1, 10000)



import random
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import scipy.stats as stats
from jinja2 import Template
from IPython.display import HTML
import base64
from io import BytesIO
import pandas as pd


from ..datahub import DataHub
from ..dtype import *


style = """                
.my-custom-styles body {
    font-family: Arial, sans-serif;
}

.container {
    padding: 10px;
    margin: 1px;
    box-sizing: border-box;
}

.flex-container {
    display: flex;
    flex-wrap: wrap;  /* Add flex-wrap for responsive layout */
}

.flexing {
    flex-grow: 1;
}

.con-100 {
    width: 10px;
}

.container-200 {
    width: 200px;  /* Set the desired width directly */
}

.container-300 {
    width: 350px;  /* Set the desired width directly */
}

.container-400 {
    width: 450px;  /* Set the desired width directly */
}

.container-500 {
    width: 488px;  /* Set the desired width directly */
}

.container-800 {
    width: 800px;  /* Set the desired width directly */
}

.container-1000 {
    width: 1000px;  /* Set the desired width directly */
}

.img-container {
    width: 100%;
    text-align: center;
}

.img-thumb {
    overflow: hidden;
    height: auto;
    width: 100%;
}

.mytable {
    /* Consider removing float: left; and width: 100%; if not needed */
    font-family: Tahoma, Algerian, Geneva, sans-serif;
    width: 100%;
    border-collapse: collapse;
}

.mytable td {
    padding: 8px;  /* Adjusted padding for better spacing */
    border: 1px solid #dddfe1;
    border-left: none;
    border-right: none;
    height: 34px;  /* Set your desired height for the table cells */
    white-space: nowrap !important;  /* Prevent text from wrapping */
}

.mytable thead td {
    text-align: left;
    background-color: #386087;
    color: #ffffff;
    font-weight: bold;
    font-size: 13px;
    border: 1px solid #54585d;
}

.mytable tbody td {
    text-align: left !important;
    color: #636363;
}

.mytable tbody tr:nth-child(even) {
    background-color: white;
}

.mytable tbody tr:nth-child(even):hover {
    background-color: #e3f2fd;
}

.mytable tbody tr td:first-child {
    font-weight: bold;
}

"""
def create_container_html(container_name, container_keys, data, cfg):
    hover_color = cfg.get('hover_color', '#e3f2fd')  # Default hover color
    
    # Set the widths based on the container_name
    con_size = '200'
    if container_name == 'Stats':
        con_size = '300'
        stat_width = '220px'  # 1/4 width
        value_width = '80px'  # 3/4 width
        
    elif container_name == 'Alerts':
        con_size = '400'
        stat_width = '370px'  # 1/4 width
        value_width = '80px'  # 3/4 width
    
    elif container_name == 'container3a':
        stat_width = '25%'  # 1/4 width
        value_width = '75%'  # 3/4 width
    else:
        stat_width = '50%'  # Default width for other containers
        value_width = '50%'


    # Check if container_name is in the specified list
    if container_name in ['Alerts', 'Stats', 'Columns']:
        # Add table headers based on the condition
         table_headers = f"<th style='width: {stat_width}; text-align: center !important; font-size: larger; color: #333;'>{container_name}</th>"
    else:
        # If container_name is not in the list, initialize an empty string for table_headers
        table_headers = ""

    table_rows = []
    # Iterate through the data items
    for stat, value in data.items():
        if stat in container_keys:
            if container_name == 'Alerts':
                # If the value is a dictionary, iterate through its items and create additional table rows
                table_rows.extend(f"<tr><td style='width: {stat_width};'>"
                                  f"<span style='color: #FF6961;'>{inner_stat}</span> has {inner_value[0]} {stat} {inner_value[1]}."
                                  #f"</td><td style='width: {value_width}; text-align: center !important;'>{inner_value[2]} </td></tr>" # Alert 
                                  for inner_stat, inner_value in value.items())
            else:
                # If the value is not a dictionary, create a regular table row
                table_rows.append(f"<tr><td style='width: {stat_width};'>{stat}</td><td style='width: {value_width}; text-align: center !important;'>{value}</td></tr>")

    return f"""
        <div class="container container-{con_size}">
            <table class="mytable">
                {table_headers}
                {"".join(table_rows)}
            </table>
        </div>
    """


def create_main_html(template_title, style, containers_html, pics, cfg):
    hover_color = cfg.get('hover_color', '#e3f2fd') #default
    img_width = "600px" if template_title == "val_tableC" else "500px"
    img_pad = "0px" if template_title == "val_tableC" else "0px"
    image_section = f"""
        <div class="container-500">
            <div class="img-container-cat" style="width:{img_width}; padding-left: {img_pad};">
                <img class="img-thumb-cat" style="" src="data:image/png;base64, {pics[0]}" alt="Histogram">
            </div>
        </div>
    """ if template_title != "val_tableW" else ""

    return f"""
        <!DOCTYPE html>
        <html>
        <head>
            <title>{template_title}</title>
            <style>{style}</style>
            <style>
                .mytable tbody tr:nth-child(even):hover {{
                    background-color: {hover_color} !important;
                }}
                .mytable tbody tr:nth-child(odd):hover {{
                    background-color: {hover_color} !important;
                }}
            </style>
        </head>
        <body class="my-custom-styles">
            <div class="flex-container" style="flex-wrap: wrap;">
                {containers_html}

                {image_section}
            </div>

            <div class="flex-container">
                <div class="container container-800">Hello</div>
                <div class="container">Con 2</div>
                <div class="container">Con 2</div>
                <div class="container">Con 2</div>
            </div>
            <!-- New container with bottom border -->
            <div class="container-gen" style="width: 100%; border-bottom: 1px solid #DDDFE1;"></div>
        </body>
        </html>
    """


def generate_table(data,container_keys,table_name, style=None, *pics, cfg=None):
    cfg = cfg or {}
    containers_html = "".join(create_container_html(container_name, container_keys, data, cfg) for container_name, container_keys in container_keys.items())
    return create_main_html(table_name, style, containers_html, pics, cfg)


##############################################################################################################################


def value_tableX(col,cfg,tokens=None, df=None):

    if tokens is not None:
        #dat = tokens
        dat = col.value_counts()
    else:
        dat = col.value_counts()


    # Separate the data into the first 20 rows and the rest
    top_20 = dat.head(10)
    other_values = dat.iloc[10:]

    # Create rows for the first 20 values
    rows_top_20 = [f"<tr><td>{idx[:20] if isinstance(idx, str) else idx}</td><td style='text-align: left !important;'>{val}</td><td>{valp}</td></tr>" for idx, val, valp in zip(top_20.index, top_20.values, np.round((top_20.values / col.size) * 100, 2))]

    # Create a row for the other values
    row_other_values = f"<tr><td>Other Values</td><td style='text-align: left !important;'>{other_values.sum()}</td><td>{np.round((other_values.sum() / col.size) * 100, 2)}</td></tr>"

    # Concatenate all rows
    all_rows = "".join(rows_top_20) + row_other_values

    top3 = col.head(3)
    low3 = col.tail(3)

    hover_color = cfg.get('hover_color','#e3f2fd')

    
    # Create a value table with f-strings (modify this as needed)
    value_table = f"""
        <!DOCTYPE html>
        <html>
        <head>
            <title>val_tableC</title>
            <style>
                .mycs body {{
                    font-family: Arial, sans-serif;
                }}

                .containerX {{
                    padding: 10px;
                    margin: 1px;
                    box-sizing: border-box;
                }}

                .flex-containerX {{
                    display: flex;
                    flex-wrap: wrap;
                }}

                .container-200X {{
                    flex: 2;
                }}

                .mytableX {{
                    float: left;
                    width: 100%;
                    font-family: Tahoma, Algerian, Geneva, sans-serif;
                    border-collapse: collapse;
                }}

                .mytableX td {{
                    padding: 8px;
                    border: 1px solid #dddfe1;
                    border-left: none;
                    border-right: none;
                    height: 34px;
                    white-space: nowrap !important;
                }}

                .mytableX thead td {{
                    text-align: left !important;
                    background-color: #ffffff; /* White background color */
                    color: #386087; /* Your desired text color */
                    font-weight: bold;
                    font-size: 13px;
                    padding: 10px; /* Adjusted padding for better spacing */
                }}

                .mytableX tbody td {{
                    text-align: left !important;
                    color: #636363;
                }}

                .mytableX tbody tr:nth-child(even) {{
                    background-color: white;
                }}

                .mytableX tbody tr:nth-child(even):hover {{
                    background-color: {hover_color} !important;
                }}

                .mytableX tbody tr:nth-child(odd):hover {{
                    background-color: {hover_color} !important;
                }}

                .mytableX tbody tr td:first-child {{
                    font-weight: bold;
                }}
            </style>
        </head>
         <body class="mycs">
            <div class="flex-containerX">
                <div class="containerX container-200X">
                   <table class="mytableX">
                        <thead>
                            <tr>
                                <td>Values</td>
                                <td>Count</td>
                                <td>Frequency (%)</td>
                            </tr>
                        </thead>
                        {all_rows}
                    </table>
                </div>
            <div class="flex-containerX">
                <div class="containerX container-200X" style="width: 600px;">
                <div>
                <table class="mytableX" style="margin-bottom: 34px !important;">
                        <thead>
                            <tr>
                                <td style="width: 150px;">Head</td>
                                <td>Values</td>
                               
                            </tr>
                        </thead>
                        {''.join(f'<tr><td style="width: 150px;">{stat}</td><td style="text-align: left;">{value}</td></tr>' for stat, value in top3.items())}
                    </table>
                     </div>
                     <table class="mytableX">
                        <thead>
                            <tr>
                                <td style="width: 150px;">Tail</td>
                                <td>Values</td>
                                
                            </tr>
                        </thead>
                        {''.join(f'<tr><td style="width: 150px;">{stat}</td><td style="text-align: left;">{value}</td></tr>' for stat, value in low3.items())}
                    </table>
                     </div>
                    </div>
                     
            </div>
            <div class="container-x" style="width: 100%; border-bottom: 1px solid #DDDFE1;"></div>
            </div>
        </body>
        </html>
    """
    return value_table

    
def value_tableW(df,cfg):
    # Convert the first and last 10 rows of the DataFrame to HTML
    df_html_head = df.head(10).to_html(index=False, justify='left', escape=False)
    df_html_tail = df.tail(10).to_html(index=False, justify='left', escape=False)

    hover_color = cfg.get('hover_color','#e3f2fd')
    num = cfg.get('w_get_rows',5)

    # Create a value table with basic styling
    value_table = f"""
        <!DOCTYPE html>
        <html>
        <head>
            <title>DataFrame Value Table</title>
            <style>
                .dfw tbody tr:nth-child(even):hover {{
                    background-color: {hover_color} !important;
                }}
                .dfw tbody tr:nth-child(odd):hover {{
                    background-color: {hover_color} !important;
                }}
                </style>
           
        </head>
        <body>
            <div class="dfw">
                <h2>Top 10 Rows</h2>
                {df_html_head}
                <h2>Bottom 10 Rows</h2>
                {df_html_tail}
            </div>
        </body>
        </html>
    """
    return value_table





import pandas as pd
import hashlib

from ..wizrenderer import WizRenderer
from ..dtype import *
from ..configs import *

from .calc import calc
from .wiz import wiz
 


def eda(df, col1=None, col2=None, vars=None, cache=None, wrap=True,stats=False):
    # Configuration settings
    cfg = Configs.current_theme

    # Get dtypes
    dtypes = Configs.get_dtypes(df)

    # Check if caching is enabled
    use_cache = Configs.use_eda_cache if cache is None else cache

    # Measure the start time
    #start_time = time.time()

    # Assuming eda_cache is the dictionary you want to get the size of
    memory_size = sys.getsizeof(Configs.eda_cache)
    
    # logger.debug(f"State of cache: {Configs.use_eda_cache}")
    # logger.debug(f"Memory size of eda_cache: {memory_size} bytes")
    # logger.debug(f"Number of cache keys: {(len(Configs.eda_cache.keys()))}")

    # Generate a hash key for caching based on function parameters
    cache_key = hashlib.sha256(f"{col1}-{col2}-{dtypes}".encode()).hexdigest()

    # Check if result is cached
    if use_cache and cache_key in Configs.eda_cache:
        cached_result = Configs.eda_cache[cache_key]
        cached_df = cached_result['df']

        # Now, compare the two DataFrames
        if col1 is None:
            are_equal = df.equals(cached_df)
        elif col1 == 'WRAPPER':
            are_equal = df.equals(cached_df)
        else:
            are_equal = df[col1].equals(cached_df[col1])
        if are_equal:
            # Return the cached result
            #logger.debug("Returning cached result.")
            #elapsed_time = time.time() - start_time
            #logger.debug(f"Execution time: {elapsed_time} seconds")
            return WizRenderer(cached_result['data_load'], cached_result['variable_type'], Configs.current_theme)

    # Calculate
    hub = calc(df, col1, col2, cfg, dtypes, vars, wrap)
    #print(hub)
    if hub is None:  # because plot(df) is recursive
        #logger.warning("Calculation result is None. its in init.py eda #calculate ")
        #elapsed_time = time.time() - start_time
        #logger.debug(f"Execution time: {elapsed_time} seconds")
        return 0

    # Render
    data_load = wiz(hub, cfg)
    
    # Save the result to cache
    if use_cache:
        Configs.eda_cache[cache_key] = {
            'data_load': data_load,
            'variable_type': hub.variable_type,
            'df': df.copy()  # Cache a copy of the DataFrame
        }
        #logger.debug("Result saved to cache.")

    # Measure the end time
    #elapsed_time = time.time() - start_time
    #logger.debug(f"Execution time: {elapsed_time} seconds")

    # # STATS
    # print('stat',stats==True)
    # if stats == True:
    #     return hub
        
    # else:
    #     return WizRenderer(data_load, hub.variable_type, cfg)

    return WizRenderer(data_load, hub.variable_type, cfg)


from .dtype import *

import sys



class Theme:

    PEACH =  {
    'dash_donuts_color': '#FFCBA5',
    'dash_bars_color': '#FF9899',
    'dash_bars_text_color': '#525252',
    'label_color': '#FF9899',
    'mini_hist_color': '#FFCBA5',
    'hist_color': '#FFCBA5',
    'hist_kde_color': '#FF9899',
    'bar_color': 'light:#FF9899',
    'bar_font_color': '#525252',
    'hover_color': '#FF9899'}

    FOREST = {
    'dash_donuts_color': '#a3b18a',
    'mini_hist_color': '#a3b18a',
    'hist_color': '#a3b18a',
    
    'label_color':'#bc6c25',
    'dash_bars_color': '#bc6c25',
    'hist_kde_color': '#bc6c25',
    'bar_color': 'light:#bc6c25',
    'hover_color': '#a3b18a',
    
    'dash_bars_text_color': '#525252',
    'bar_font_color': '#525252',}

    AMERICANA = {
    'dash_donuts_color': '#c1121f',
    'mini_hist_color': '#c1121f',
    'hist_color': '#c1121f',
    
    'label_color':'#457b9d',
    'dash_bars_color': '#457b9d',
    'hist_kde_color': '#457b9d',
    'bar_color': 'light:#457b9d',
    'hover_color': '#fdf0d5',
    
    'dash_bars_text_color': '#525252',
    'bar_font_color': '#525252',}

    BABY = {
    'dash_donuts_color': '#a2d2ff',
    'mini_hist_color': '#a2d2ff',
    'hist_color': '#a2d2ff',
    
    'label_color':'#ff8fab',
    'dash_bars_color': '#ff8fab',
    'hist_kde_color': '#ff8fab',
    'bar_color': 'light:#ff8fab',
    'hover_color': '#ff8fab',
    
    'dash_bars_text_color': '#525252',
    'bar_font_color': '#525252',}

    PASTEL = {
    'dash_donuts_color': '#e4c1f9',
    'mini_hist_color': '#d0f4de',
    'hist_color': '#a9def9',
    
    'label_color':'#a9def9',
    'dash_bars_color': '#ff99c8',
    'hist_kde_color': '#e4c1f9',
    'bar_color': 'light:#e4c1f9',
    'hover_color': '#fcf6bd',
    
    'dash_bars_text_color': '#525252',
    'bar_font_color': '#525252',}

    ITALY = {
    'dash_donuts_color': '#e5383b',
    'mini_hist_color': '#e5383b',
    'hist_color': '#f07167',
    
    'label_color':'#55a630',
    'dash_bars_color': '#55a630',
    'hist_kde_color': '#55a630',
    'bar_color': 'light:#55a630',
    'hover_color': '#55a630',
    
    'dash_bars_text_color': '#525252',
    'bar_font_color': '#525252',}

    OCEAN = {
    'dash_donuts_color': '#fed9b7',
    'mini_hist_color': '#fed9b7',
    'hist_color': '#fed9b7',
    
    'label_color':'#00afb9',
    'dash_bars_color': '#00afb9',
    'hist_kde_color': '#00afb9',
    'bar_color': 'light:#00afb9',
    'hover_color': '#00afb9',
    
    'dash_bars_text_color': '#525252',
    'bar_font_color': '#525252',}

    PLATINUM = {
    'dash_donuts_color': '#c6c5b9',
    'mini_hist_color': '#c6c5b9',
    'hist_color': '#c6c5b9',
    
    'label_color':'#393d3f',
    'dash_bars_color': '#393d3f',
    'hist_kde_color': '#393d3f',
    'bar_color': 'light:#393d3f',
    'hover_color': '#bcb8b1',
    
    'dash_bars_text_color': '#525252',
    'bar_font_color': '#525252',}

    GOLD = {
    'dash_donuts_color': '#ffc43d',
    'mini_hist_color': '#ffc43d',
    'hist_color': '#ffc43d',
    
    'label_color':'#f4d35e',
    'dash_bars_color': '#f4d35e',
    'hist_kde_color': '#0b132b',
    'bar_color': 'light:#f4d35e',
    'hover_color': '#f4d35e',
    
    'dash_bars_text_color': '#525252',
    'bar_font_color': '#525252',}
    

class Configs:

    # cache flag
    use_eda_cache = True
    
    
    # Simple cache to store previously generated plots
    eda_cache = {}

    @classmethod
    def reset_cache(cls, df):
        # Iterate over keys in the cache and remove entries that correspond to the given DataFrame
        keys_to_remove = [key for key in cls.eda_cache.keys() if cls.eda_cache[key]['df'].equals(df)]
        for key in keys_to_remove:
            del cls.eda_cache[key]

        print(f"""Cache reset for DataFrame with shape {df.shape}\n
                number of cache keyslen{(Configs.eda_cache.keys())}"""            
                     )





    ################################ setting DTYPES #####################################################

    
    dtypes_cache = {}
    dtype = None

    use_id = True
    
    @classmethod
    def get_dtypes(cls, df):
        """Returns a dict of suggested dtypes for each variable"""
        if cls.use_id:
            df_name = id(df)
        else:
            df_name = str(df)
        
        cls.dtype = cls.dtypes_cache.get(df_name,None)   
        
        if cls.dtype is not None:
            return cls.dtypes_cache[df_name]
        else:
            cls.dtypes_cache[df_name] = {i: determine_variable_type(df[i])[0] for i in df.columns}       
        return cls.dtypes_cache[df_name]


    @classmethod
    def set_dtypes(cls,df, dtypes_dict=None):
        """Updates the dtypes attribute with the provided dictionary"""
        if cls.use_id:
            df_name = id(df)
        else:
            df_name = str(df)

        if dtypes_dict == 'reset':
            cls.dtypes_cache[df_name] = None
               
        cls.dtype = cls.dtypes_cache.get(df_name,None)
        
        if cls.dtype is None or dtypes_dict is None:
            cls.dtype = cls.get_dtypes(df)  # Initialize dtypes using get_dtypes if it's None
        else:
            for key, new_value in dtypes_dict.items():
                if key in cls.dtype and cls.dtype[key] != new_value:
                    cls.dtype[key] = new_value
                    cls.dtypes_cache[df_name] = cls.dtype
                    #print(cls.dtype)

        return 'done'



    
    ################################ THEME #####################################################
    
    current_theme = Theme.PEACH

    @classmethod
    def get_theme(cls, current=False):
        """
        List available themes and their values.

        Parameters:
        - values (bool, optional): If True, returns a dictionary with theme names and values.
          If False, returns a list of theme names only.

        Returns:
        - dict or list: If values is True, a dictionary of theme names and their values.
          If values is False, a list of theme names.
          
        """
        if current:
            return cls.current_theme.copy()
            # themes = {}
            # for attr in cls.current_theme:
            #     if not callable(getattr(Theme, attr)) and not attr.startswith("__"):
            #         themes[attr] = getattr(Theme, attr)
            # return themes
        else:
            return [attr for attr in dir(Theme) if not callable(getattr(Theme, attr)) and not attr.startswith("__")]

    @classmethod
    def set_theme(cls, theme_name):
        """Set the current theme based on the provided theme name."""
        if hasattr(Theme, theme_name):
            cls.use_eda_cache = False
            cls.eda_cache = {}
            cls.current_theme = getattr(Theme, theme_name)
            cls.use_eda_cache = True
            
            
        else:
            print(f"Theme '{theme_name}' not found. Using the default theme.")

    @classmethod
    def add_theme(cls, theme_name, theme_values):
        """
        Add a new theme to the Theme class.
    
        Parameters:
        - cls: The Theme class.
        - theme_name (str): The name of the new theme.
        - theme_values (dict): A dictionary containing the values of the new theme.
    
        Returns:
        - None
    
        Example:
        >>> Configs.add_theme('NEW_THEME', {'hist_color': 'purple', 'tabset': '#FFA500'})
        """
        setattr(Theme, theme_name, theme_values)


# setup.py
#

from setuptools import setup, find_packages

from pathlib import Path

source_root = Path(".")
with (source_root / "README.md").open(encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="adix",
    version="0.1.0a",
    author='Marek Jindrich',
    author_email='adix.eda@gmail.com',
    description='Automated exploratory data analysis (EDA).',
    python_requires='>=3.10',
    packages=find_packages(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Environment :: Console",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Healthcare Industry",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Framework :: IPython",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    install_requires=[
        'pandas>=2.1.0',
        'numpy>=1.26.2',
        'matplotlib>=3.8.2',
        'seaborn>=0.13.0',
        'scipy>=1.11.4',
        'Jinja2>=3.1.2',
        'notebook>=7.0.6',
        'wordcloud>=1.8.1',
       	'nltk >= 3.6.7'
    ],
    keywords="pandas eda data data-science data-analysis python jupyter ipython",
    long_description=long_description,
    long_description_content_type="text/markdown",
    
)



import random
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import scipy.stats as stats
from jinja2 import Template
from IPython.display import HTML
import base64
from io import BytesIO
import pandas as pd

from wordcloud import WordCloud


from ..datahub import DataHub
from ..dtype import *

from .tables import *


"""""" # NEW IMPORTS
from nltk.tokenize import word_tokenize
""""""


"""
This module implements the visualization for the plot(df) function.
"""  


        
# def hist_plot_thumb_mini(data,cfg, num_bins=None):

#     # Exclude NaN values
#     data_non_nan = data.dropna()
    
#     unique = data.nunique()
#     fig_histogram, ax_histogram = plt.subplots(figsize=(3, 0.4))

#     if num_bins is None:
#         num_bins = min(unique, 30)  # Adjust 40 to your preference or remove for automatic calculation

#     ax_histogram.hist(data_non_nan, bins=num_bins, color=cfg['mini_hist_color'], edgecolor='white') #color='#06c'
#     ax_histogram.yaxis.set_visible(False)
#     ax_histogram.set_yscale('log')  # Set logarithmic scale on the y-axis TO HANDLE VISUALIZE OUTLIERS!!!
#     ax_histogram.xaxis.set_visible(False)
#     ax_histogram.set_frame_on(False)
#     plt.grid(False)

#     # Save each figure to a buffer
#     buffer_histogram = BytesIO()
#     fig_histogram.savefig(buffer_histogram, format='png')
#     image_histogram_thumb_mini = base64.b64encode(buffer_histogram.getvalue()).decode('utf-8')
#     plt.close(fig_histogram)

#     return image_histogram_thumb_mini

""" POSSIBLE IMPORVEMENT calling plt.hist instad of plt.subplots """

def hist_plot_thumb_mini(data, cfg, num_bins=None):
    # Exclude NaN values
    data_non_nan = data.dropna()

    unique = data.nunique()

    if num_bins is None:
        num_bins = min(unique, 30)

    plt.figure(figsize=(3, 0.4))
    plt.hist(data_non_nan, bins=num_bins, color=cfg['mini_hist_color'], edgecolor='white')
    plt.yscale('log')
    plt.axis('off')
    plt.grid(False)

    buffer_histogram = BytesIO()
    plt.savefig(buffer_histogram, format='png')
    image_histogram_thumb_mini = base64.b64encode(buffer_histogram.getvalue()).decode('utf-8')
    plt.close()

    return image_histogram_thumb_mini
    
   
def hist_plot(data,cfg):
        fig_histogram, ax_histogram = plt.subplots(figsize=(6, 4))
        ax_histogram.hist(data, bins=30, color='skyblue', edgecolor='black')
        ax_histogram.set_title('Histogram')
        ax_histogram.set_xlabel('Value')
        ax_histogram.set_ylabel('Frequency')
    
        # Save each figure to a buffer
        buffer_histogram = BytesIO()
        fig_histogram.savefig(buffer_histogram, format='png')
        image_histogram = base64.b64encode(buffer_histogram.getvalue()).decode('utf-8')
        plt.close(fig_histogram)

        return image_histogram
'''testing hist + kde '''

# def hist_plot_thumb(data,cfg, name=None):
#     # Create a histogram
#     fig_histogram, ax_histogram = plt.subplots(figsize=(6, 4))
#     ax_histogram.hist(data, bins=30, color=cfg['hist_color'], edgecolor='white', density=True) #'skyblue'
    
#     # Add a kernel density plot
#     sns.kdeplot(data, color=cfg['hist_kde_color'], ax=ax_histogram, warn_singular=False) #'pink'
    
#     ax_histogram.tick_params(axis='both', colors='#a9a9a9')
#     ax_histogram.set_xlabel('Value', color='#a9a9a9')
#     ax_histogram.set_ylabel('Frequency', color='#a9a9a9')
#     if name is not None:
#         ax_histogram.set_title(name)
#     ax_histogram.set_frame_on(False)
#     plt.grid(False)

#     # Save the figure to a buffer
#     buffer_histogram = BytesIO()
#     fig_histogram.savefig(buffer_histogram, format='png')
#     image_histogram_thumb = base64.b64encode(buffer_histogram.getvalue()).decode('utf-8')
#     plt.close(fig_histogram)

#     return image_histogram_thumb

    
def kde_plot(data):
        fig_kde, ax_kde = plt.subplots(figsize=(6, 4))
        sns.kdeplot(data, color='salmon')
        ax_kde.set_title('KDE Plot')
        ax_kde.set_xlabel('Value')
        ax_kde.set_ylabel('Density')

        # Save figure to a buffer
        buffer_kde = BytesIO()
        fig_kde.savefig(buffer_kde, format='png')
        image_kde = base64.b64encode(buffer_kde.getvalue()).decode('utf-8')
        plt.close(fig_kde)

        return image_kde
#axxx    cfg['mini_hist_color']    
def qq_plot(data, cfg):
    # Exclude NaN values
    data_non_nan = data.dropna()

    # Set the plot border to light grey
    fig_qq, ax_qq = plt.subplots(figsize=(6, 4))
    
    qq_line = stats.probplot(data_non_nan, plot=ax_qq, fit=True, rvalue=True)
    
    # Customize plot colors based on cfg dictionary
    line_color = cfg.get('blue', 'blue')
    marker_color = cfg.get('mini_hist_color', 'red')
    
    # Customize line and marker colors
    ax_qq.get_lines()[0].set_markerfacecolor(marker_color)
    ax_qq.get_lines()[0].set_markeredgecolor(marker_color)
    ax_qq.get_lines()[0].set_color(line_color)
    
    # Customize text colors
    title_color = cfg.get('title_color', '#a9a9a9')
    xlabel_color = cfg.get('xlabel_color', '#a9a9a9')
    ylabel_color = cfg.get('ylabel_color', '#a9a9a9')
    tick_color = cfg.get('tick_color', '#a9a9a9')

    ax_qq.set_title('Normal Q-Q Plot', color=title_color)
    ax_qq.set_xlabel('Theoretical Quantiles', color=xlabel_color)
    ax_qq.set_ylabel('Ordered Values', color=ylabel_color)

    # Set spines color to light grey
    for spine in ax_qq.spines.values():
        spine.set_edgecolor('#d3d3d3')

    # Set xticks and yticks color to #a9a9a9
    ax_qq.tick_params(axis='x', colors=tick_color)
    ax_qq.tick_params(axis='y', colors=tick_color)
    ax_qq.grid(color=tick_color, linestyle=':', linewidth=0.5, alpha=0.5)

    # Save figure to a buffer
    buffer_qq = BytesIO()
    fig_qq.savefig(buffer_qq, format='png')
    image_qq = base64.b64encode(buffer_qq.getvalue()).decode('utf-8')
    plt.close(fig_qq)

    return image_qq


def box_plot(data):
        fig_box, ax_box = plt.subplots(figsize=(6, 4))
        sns.boxplot(data, color='lightgreen')
        ax_box.set_title('Box Plot')
        ax_box.set_xlabel('Value')

        # Save figure to a buffer
        buffer_box = BytesIO()
        fig_box.savefig(buffer_box, format='png')
        image_box = base64.b64encode(buffer_box.getvalue()).decode('utf-8')
        plt.close(fig_box)

        return image_box

def bar_chart(data):
    """
    Generate a horizontal bar chart.

    Parameters:
    - data (pandas.Series or pandas.DataFrame): Data for the bar chart.

    Returns:
    - str: Base64-encoded PNG image of the bar chart.
    """
    if isinstance(data, pd.DataFrame):
        data = data.squeeze()  # Convert DataFrame to Series if needed

    value_counts = data.value_counts(normalize=True) * 100
    
    fig_bar, ax_bar = plt.subplots(figsize=(6, 4))
    value_counts.plot(kind='barh', color='skyblue', edgecolor='black', ax=ax_bar)
    
    ax_bar.set_title('Horizontal Bar Chart')
    ax_bar.set_xlabel('Count')
    ax_bar.set_ylabel('Category')

    # Set explicit ticks and add percentage signs
    ticks = ax_bar.get_xticks()
    tick_labels = [f'{tick:.2f}%' for tick in ticks]
    ax_bar.set_xticks(ticks)
    ax_bar.set_xticklabels(tick_labels)

    
    # Save figure to a buffer
    buffer_bar = BytesIO()
    fig_bar.savefig(buffer_bar, format='png')
    image_bar = base64.b64encode(buffer_bar.getvalue()).decode('utf-8')
    plt.close(fig_bar)

    return image_bar

###########################################################################################################################
######################################### BIVARIATE CHARTS ###########################################################
###########################################################################################################################

"""BIVARIATE"""
# #axxx



def scatter_plot(x, y):
    # Create a DataFrame from the data
    df = pd.DataFrame({'X': x, 'Y': y})

    # Create a scatter plot
    fig_scatter, ax_scatter = plt.subplots(figsize=(8, 6))
    sns.scatterplot(x='X', y='Y', data=df, color='coral', alpha=0.6, ax=ax_scatter)

    # Set plot title and labels
    ax_scatter.set_title('Scatter Plot', color='#a9a9a9')
    ax_scatter.set_xlabel('X', color='#a9a9a9')
    ax_scatter.set_ylabel('Y', color='#a9a9a9')

    # Add grid
    ax_scatter.grid(color='#a9a9a9', linestyle='dashed', linewidth=0.5)

    # Set border color of the plot to light grey
    for spine in ax_scatter.spines.values():
        spine.set_edgecolor('lightgrey')

    # Set font color for all text in the chart
    for text in ax_scatter.texts + ax_scatter.get_xticklabels() + ax_scatter.get_yticklabels():
        text.set_color('#a9a9a9')

    # Save figure to a buffer
    buffer_scatter = BytesIO()
    fig_scatter.savefig(buffer_scatter, format='png')
    image_scatter = base64.b64encode(buffer_scatter.getvalue()).decode('utf-8')
    plt.close(fig_scatter)

    return image_scatter

def hexbin_plot(x, y):
    # Create a hexbin plot
    fig_hexbin, ax_hexbin = plt.subplots(figsize=(8, 6))
    hb = ax_hexbin.hexbin(x, y, gridsize=30, cmap='Blues')

    # Set plot title and labels
    ax_hexbin.set_title('Hexbin Plot', color='#a9a9a9')
    ax_hexbin.set_xlabel('X', color='#a9a9a9')
    ax_hexbin.set_ylabel('Y', color='#a9a9a9')

    # Add colorbar
    cbar = plt.colorbar(hb)
    cbar.set_label('Count', color='#a9a9a9')

    # Set font color for ticks on both x and y axes
    ax_hexbin.tick_params(axis='both', colors='#a9a9a9')

    # Set border color of the plot to light grey
    for spine in ax_hexbin.spines.values():
        spine.set_edgecolor('lightgrey')

    # Save figure to a buffer
    buffer_hexbin = BytesIO()
    fig_hexbin.savefig(buffer_hexbin, format='png')
    image_hexbin = base64.b64encode(buffer_hexbin.getvalue()).decode('utf-8')
    plt.close(fig_hexbin)

    return image_hexbin


# def scatter_plot(x, y):
#     fig_scatter, ax_scatter = plt.subplots(figsize=(8, 6))
#     ax_scatter.scatter(x, y, color='coral', alpha=0.6)
#     ax_scatter.set_title('Scatter Plot')
#     ax_scatter.set_xlabel('X')
#     ax_scatter.set_ylabel('Y')

#     # # Add grid
#     # ax_scatter.grid(True)

#     # Save figure to a buffer
#     buffer_scatter = BytesIO()
#     fig_scatter.savefig(buffer_scatter, format='png')
#     image_scatter = base64.b64encode(buffer_scatter.getvalue()).decode('utf-8')
#     plt.close(fig_scatter)

#     return image_scatter

# def hexbin_plot(x, y):
#     fig_hexbin, ax_hexbin = plt.subplots(figsize=(8, 6))
#     ax_hexbin.hexbin(x, y, gridsize=30, cmap='Blues')
#     ax_hexbin.set_title('Hexbin Plot')
#     ax_hexbin.set_xlabel('X')
#     ax_hexbin.set_ylabel('Y')

#     # Save figure to a buffer
#     buffer_hexbin = BytesIO()
#     fig_hexbin.savefig(buffer_hexbin, format='png')
#     image_hexbin = base64.b64encode(buffer_hexbin.getvalue()).decode('utf-8')
#     plt.close(fig_hexbin)

#     return image_hexbin
    
""" doesnt work check"""

# def pair_plot(dataframe):
#     fig_pair, ax_pair = plt.subplots(figsize=(8, 8))
#     sns.pairplot(dataframe, palette='viridis', diag_kind='kde', markers='o')
#     ax_pair.set_title('Pair Plot')

#     # Save figure to a buffer
#     buffer_pair = BytesIO()
#     fig_pair.savefig(buffer_pair, format='png')
#     image_pair = base64.b64encode(buffer_pair.getvalue()).decode('utf-8')
#     plt.close(fig_pair)

#     return image_pair

# def correlation_heatmap(dataframe):
#     fig_heatmap, ax_heatmap = plt.subplots(figsize=(8, 6))
#     sns.heatmap(dataframe.corr(), annot=True, cmap='coolwarm', fmt=".2f")
#     ax_heatmap.set_title('Correlation Heatmap')

#     # Save figure to a buffer
#     buffer_heatmap = BytesIO()
#     fig_heatmap.savefig(buffer_heatmap, format='png')
#     image_heatmap = base64.b64encode(buffer_heatmap.getvalue()).decode('utf-8')
#     plt.close(fig_heatmap)

#     return image_heatmap
#ax
# def box_plot_con(x, y):

#     # Ensure x and y are numeric (convert if necessary)
#     fig_box, ax_box = plt.subplots(figsize=(8, 6))
#     sns.boxplot(x=x, y=y, ax=ax_box)
    
#     ax_box.set_title('Box Plot: Numerical vs Categorical')
#     ax_box.set_xlabel('Categorical Variable')
#     ax_box.set_ylabel('Numerical Variable')

#     # Save figure to a buffer
#     buffer_box = BytesIO()
#     fig_box.savefig(buffer_box, format='png')
#     image_box_con = base64.b64encode(buffer_box.getvalue()).decode('utf-8')
#     plt.close(fig_box)

#     return image_box_con

def box_plot_con(x, y):
    # Create a DataFrame from the data
    df = pd.DataFrame({'X': x, 'Y': y})

    # Create a box plot
    fig_box, ax_box = plt.subplots(figsize=(8, 6))
    sns.boxplot(x='X', y='Y', data=df, hue='X', palette='Set3', ax=ax_box, dodge='Auto', legend=False)

    # Set plot title and labels
    ax_box.set_title('Box Plot', color='#a9a9a9')
    ax_box.set_xlabel(x.name, color='#a9a9a9')
    ax_box.set_ylabel(y.name, color='#a9a9a9')

    # Set x-axis ticks and labels
    x_ticks = np.arange(len(ax_box.get_xticklabels()))
    ax_box.set_xticks(x_ticks)
    
    # Crop x-axis labels if longer than 20 symbols
    labels = [text.get_text() if len(text.get_text()) <= 10 else text.get_text()[:10] + '...' for text in ax_box.get_xticklabels()]
    ax_box.set_xticklabels(labels, ha='center')

    # Set font color for all text in the chart
    for text in ax_box.texts + ax_box.get_xticklabels() + ax_box.get_yticklabels():
        text.set_color('#a9a9a9')

    ax_box.xaxis.grid(color='#a9a9a9', linestyle='dashed', linewidth=0.5)
    ax_box.yaxis.grid(color='#a9a9a9', linestyle='dashed', linewidth=0.5)

    # Set the plot border color to light grey
    ax_box.spines['top'].set_color('#d3d3d3')
    ax_box.spines['bottom'].set_color('#d3d3d3')
    ax_box.spines['left'].set_color('#d3d3d3')
    ax_box.spines['right'].set_color('#d3d3d3')

    # Save figure to a buffer
    buffer_box = BytesIO()
    fig_box.savefig(buffer_box, format='png')
    image_box = base64.b64encode(buffer_box.getvalue()).decode('utf-8')
    plt.close(fig_box)

    return image_box

def violin_plot(x, y):
    # Create a DataFrame from the data
    df = pd.DataFrame({'X': x, 'Y': y})

    # Create a violin plot
    fig_violin, ax_violin = plt.subplots(figsize=(8, 6))
    sns.violinplot(x='X', y='Y', data=df, hue='X', palette='Set3', ax=ax_violin, inner='quartile', split=False, legend=False)

    # Set plot title and labels
    ax_violin.set_title('Violin Plot', color='#a9a9a9')
    ax_violin.set_xlabel(x.name, color='#a9a9a9')
    ax_violin.set_ylabel(y.name, color='#a9a9a9')

    # Set x-axis ticks and labels
    x_ticks = np.arange(len(ax_violin.get_xticklabels()))
    ax_violin.set_xticks(x_ticks)
    
    # Crop x-axis labels if longer than 20 symbols
    labels = [text.get_text() if len(text.get_text()) <= 10 else text.get_text()[:10] + '...' for text in ax_violin.get_xticklabels()]
    ax_violin.set_xticklabels(labels, ha='center')

    # Set font color for all text in the chart
    for text in ax_violin.texts + ax_violin.get_xticklabels() + ax_violin.get_yticklabels():
        text.set_color('#a9a9a9')

    ax_violin.xaxis.grid(color='#a9a9a9', linestyle='dashed', linewidth=0.5)
    ax_violin.yaxis.grid(color='#a9a9a9', linestyle='dashed', linewidth=0.5)

    # Set the plot border color to light grey
    ax_violin.spines['top'].set_color('#d3d3d3')
    ax_violin.spines['bottom'].set_color('#d3d3d3')
    ax_violin.spines['left'].set_color('#d3d3d3')
    ax_violin.spines['right'].set_color('#d3d3d3')

    # Save figure to a buffer
    buffer_violin = BytesIO()
    fig_violin.savefig(buffer_violin, format='png')
    image_violin = base64.b64encode(buffer_violin.getvalue()).decode('utf-8')
    plt.close(fig_violin)

    return image_violin

# def line_chart(x, y):
#     fig_line, ax_line = plt.subplots(figsize=(6, 4))
#     ax_line.plot(x, y, marker='o', color='purple', linestyle='-', linewidth=2, markersize=8)
#     ax_line.set_title('Line Chart')
#     ax_line.set_xlabel('X')
#     ax_line.set_ylabel('Y')

#     # Save figure to a buffer
#     buffer_line = BytesIO()
#     fig_line.savefig(buffer_line, format='png')
#     image_line = base64.b64encode(buffer_line.getvalue()).decode('utf-8')
#     plt.close(fig_line)

#     return image_line


def line_chart(x, y):
    # Create a DataFrame
    df = pd.DataFrame({'Age': y, 'Sex': x})
    
    # Plot the value_counts of age separated by sex
    plt.figure(figsize=(10, 6))
    df.groupby(['Age', 'Sex']).size().unstack().plot(kind='line', linewidth=1)
    plt.title('Value Counts of Age Separated by Sex')
    plt.xlabel('Age')
    plt.ylabel('Count')
    plt.legend(title='Sex')
    plt.grid(True)

    # Save figure to a buffer
    buffer_age_distribution = BytesIO()
    plt.savefig(buffer_age_distribution, format='png')
    image_age_distribution = base64.b64encode(buffer_age_distribution.getvalue()).decode('utf-8')
    plt.close()

    return image_age_distribution


def heatmap_plot(x, y):
    # Create a DataFrame from the categorical variables
    df = pd.DataFrame({'X': x, 'Y': y})

    # Create a pivot table
    pivot_table = df.pivot_table(index='Y', columns='X', aggfunc=len, fill_value=0)

    # Create a heatmap
    plt.figure(figsize=(8, 6))
    sns.heatmap(pivot_table, annot=True, cmap='YlGnBu', fmt='g')
    plt.title('Heatmap Plot')
    plt.xlabel('X')
    plt.ylabel('Y')

    # Save figure to a buffer
    buffer_heatmap = BytesIO()
    plt.savefig(buffer_heatmap, format='png')
    image_heatmap = base64.b64encode(buffer_heatmap.getvalue()).decode('utf-8')
    plt.close()

    return image_heatmap

def stacked_bar_chart_percentage(x, y, width, height):
    # Create a DataFrame from the categorical variables
    df = pd.DataFrame({'X': x, 'Y': y})

    # Create a cross-tabulation
    cross_tab = pd.crosstab(df['X'], df['Y'])

    # Normalize the data to get percentages
    cross_tab_percentage = cross_tab.div(cross_tab.sum(axis=1), axis=0) * 100

    # Crop x-axis ticks if longer than 20 symbols
    labels = [text if len(text) <= 10 else text[:10] + '...' for text in cross_tab_percentage.index]

    # Create a stacked bar chart with percentages
    fig, ax = plt.subplots(figsize=(width, height))
    cross_tab_percentage.plot(kind='bar', stacked=True, colormap='Set3', ax=ax)

    # Rotate x ticks 45 degrees
    ax.set_xticklabels(labels, ha='center', rotation=0)

    # Set font color to a9a9a9 for all text in the chart
    for text in ax.texts + ax.get_xticklabels() + ax.get_yticklabels():
        text.set_color('#a9a9a9')

    # Set frame and grid color to a9a9a9
    for spine in ax.spines.values():
        spine.set_edgecolor('#d3d3d3')

    # ax.xaxis.grid(color='#a9a9a9', linestyle='dashed', linewidth=0.5)
    # ax.yaxis.grid(color='#a9a9a9', linestyle='dashed', linewidth=0.5)
    #plt.grid('minor')

    ax.set_title('Stacked Bar Chart (Percentage)', color='#a9a9a9')
    ax.set_xlabel('')
    ax.set_ylabel('Percentage', color='#a9a9a9')

    # Set font color for legend
    legend = ax.legend(title='Y')
    legend.get_title().set_color('#a9a9a9')

    # Save figure to a buffer
    buffer_stacked_bar_percentage = BytesIO()
    fig.savefig(buffer_stacked_bar_percentage, format='png')
    image_stacked_bar_percentage = base64.b64encode(buffer_stacked_bar_percentage.getvalue()).decode('utf-8')
    plt.close()

    return image_stacked_bar_percentage


def nested_bar_chart(x, y, width, height):
    # Create a DataFrame from the categorical variables
    df = pd.DataFrame({'X': x, 'Y': y})

    # Create a count plot
    fig, ax = plt.subplots(figsize=(width, height))
    sns.countplot(x='X', hue='Y', data=df, palette='Set3', ax=ax)


    ax.set_title('Nested Bar Chart', color='#a9a9a9')
    ax.set_xlabel('')
    ax.set_ylabel('Count', color='#a9a9a9')

    # Set font color for legend
    legend = ax.legend(title='Y')
    legend.get_title().set_color('#a9a9a9')

    # Crop x-axis ticks if longer than 20 symbols
    labels = [text.get_text() if len(text.get_text()) <= 10 else text.get_text()[:10] + '...' for text in ax.get_xticklabels()]
    ax.set_xticks(ax.get_xticks())  # Set the tick positions
    ax.set_xticklabels(labels, ha='center')

    # Set font color for all text in the chart
    for text in ax.texts + ax.get_xticklabels() + ax.get_yticklabels():
        text.set_color('#a9a9a9')

    # Set frame and grid color to a9a9a9
    for spine in ax.spines.values():
        spine.set_edgecolor('#d3d3d3')
    #ax.grid(color='#e6e6e6')

    # Save figure to a buffer
    buffer_nested_bar = BytesIO()
    fig.savefig(buffer_nested_bar, format='png')
    image_nested_bar = base64.b64encode(buffer_nested_bar.getvalue()).decode('utf-8')
    plt.close()

    return image_nested_bar

# def stacked_bar_chart_percentage(x, y,*k):
#     # Create a DataFrame from the categorical variables
#     df = pd.DataFrame({'X': x, 'Y': y})

#     # Create a cross-tabulation
#     cross_tab = pd.crosstab(df['X'], df['Y'])

#     # Normalize the data to get percentages
#     cross_tab_percentage = cross_tab.div(cross_tab.sum(axis=1), axis=0) * 100

#     # Prettify the style
#     sns.set(style="whitegrid", rc={"grid.color": "#f0f0f0", "axes.edgecolor": "#f0f0f0", "axes.linewidth": 1.0})
#     plt.figure(figsize=(8, 6))

#     # Plotting the stacked bar chart with percentages
#     colors = sns.color_palette("pastel")  # You can change the palette as needed
#     ax = cross_tab_percentage.plot(kind='bar', stacked=True, color=colors, edgecolor='none')

#     # Rotate x-axis tick labels by 90 degrees and crop if longer than 20 symbols
#     labels = [str(cls) for cls in cross_tab_percentage.index]
#     ax.set_xticklabels(labels, rotation=0, ha='center')

#     # Set the text color for ylabel, legend, and title to light grey
#     ax.yaxis.label.set_color('#a9a9a9')
#     ax.legend(title='Y', loc='upper right').get_title().set_color('#a9a9a9')
#     ax.set_title('Stacked Bar Chart (Percentage)', color='#a9a9a9')

#     # Set the text color for x-axis and y-axis tick labels to light grey
#     ax.tick_params(axis='x', colors='#a9a9a9')
#     ax.tick_params(axis='y', colors='#a9a9a9')

#     plt.xlabel(None)
#     plt.ylabel('Percentage', color='#a9a9a9')
#     #plt.grid(False)

#     # Save figure to a buffer
#     buffer_stacked_bar_percentage = BytesIO()
#     plt.savefig(buffer_stacked_bar_percentage, format='png')
#     image_stacked_bar_percentage = base64.b64encode(buffer_stacked_bar_percentage.getvalue()).decode('utf-8')
#     plt.close()

#     return image_stacked_bar_percentage




# def nested_bar_chart(x, y,*k):
#     # Create a DataFrame from the categorical variables
#     df = pd.DataFrame({'X': x, 'Y': y})

#     # Prettify the style
#     sns.set(style="whitegrid", rc={"grid.color": "#f0f0f0", "axes.edgecolor": "#f0f0f0", "axes.linewidth": 1.0})
#     plt.figure(figsize=(8, 6))

#     # Create a count plot
#     sns.countplot(x='X', hue='Y', data=df, palette='Set3', edgecolor='none')

#     plt.title('Nested Bar Chart', color='#a9a9a9')
#     plt.xlabel(None)
#     plt.ylabel('Count', color='#a9a9a9')
    
#     # Get legend title and set its color
#     legend = plt.legend(title='Y')
#     legend.get_title().set_color('#a9a9a9')

#     # Set the legend text color to light grey
#     for text in legend.get_texts():
#         text.set_color('#a9a9a9')

#     # Set the text color for x-axis tick labels to light grey
#     plt.xticks(color='#a9a9a9')
#     plt.yticks(color='#a9a9a9')

#     # Save figure to a buffer
#     buffer_nested_bar = BytesIO()
#     plt.savefig(buffer_nested_bar, format='png')
#     image_nested_bar = base64.b64encode(buffer_nested_bar.getvalue()).decode('utf-8')
#     plt.close()

#     return image_nested_bar


#abab


###########################################################################################################################
######################################### TEXT CHARTS ###########################################################
###########################################################################################################################



def word_cloud(data, cfg, name=None):
    # Combine all text in the specified column
    text = ' '.join(data.astype(str))
    
    cfg = {
    'width': 300,
    'height': 200,
    'fig_width': 5,
    'fig_height': 3.334,
    'max_words': 50,
    }

    # Generate the word cloud
    try:
        wordcloud = WordCloud(width=cfg['width'], height=cfg['height'], background_color='white',max_words=cfg['max_words']).generate(text)
    except ValueError:
        wordcloud = WordCloud(width=cfg['width'], height=cfg['height'], background_color='white',max_words=cfg['max_words']).generate('untokenable')
    # Access the tokenized words using the process_text method
    #tokens = wordcloud.process_text(text)
    tokens = word_tokenize(text)

    # Plot the WordCloud image
    fig_wordcloud, ax_wordcloud = plt.subplots(figsize=(cfg['fig_width'], cfg['fig_height']))
    ax_wordcloud.imshow(wordcloud, interpolation='bilinear')
    ax_wordcloud.axis('off')  # Turn off axis labels
    if name is not None:
        ax_wordcloud.set_title(name)
    ax_wordcloud.set_frame_on(False)
    plt.grid(False)

    # Save the figure to a buffer
    buffer_wordcloud = BytesIO()
    fig_wordcloud.savefig(buffer_wordcloud, format='png')
    image_wordcloud_thumb = base64.b64encode(buffer_wordcloud.getvalue()).decode('utf-8')
    plt.close(fig_wordcloud)

    return image_wordcloud_thumb, tokens

# Example usage
# Assuming you have a DataFrame named 'df' with a column 'text_column'






###########################################################################################################################
######################################### CATEGORICAL CHARTS ###########################################################
###########################################################################################################################


def bar_chart_thumb(data, cfg, threshold=0.15):
    """
    Generate a horizontal bar chart.

    Parameters:
    - data (pandas.Series or pandas.DataFrame): Data for the bar chart.
    - threshold (float): Proportion threshold to decide whether to place the count inside or outside the bar.

    Returns:
    - str: Base64-encoded PNG image of the bar chart.
    """
    if isinstance(data, pd.DataFrame):
        data = data.squeeze()  # Convert DataFrame to Series if needed

    value_counts = data.value_counts()

    # Check if there are more than 5 unique values
    if len(value_counts) > 5:
        # Combine smaller categories into 'others'
        others_count = value_counts.iloc[5:].sum()
        value_counts = pd.concat([value_counts.head(5), pd.Series([others_count], index=['others'])])

    total_count = value_counts.sum()
    

    # Calculate a proportional width based on the number of bars
    max_width = 0.8
    width = max_width if len(value_counts) > 2 else 0.4
    length_factor = 1.5  # Adjust this factor to control the influence of name length

    #'bar_color': 'light:#474B7B'
    col_in = cfg['bar_color']
    colors = sns.color_palette(col_in)
    
    fig_bar, ax_bar = plt.subplots(figsize=(8, 3))  # Customizable width and height
    bars = value_counts.plot(kind='barh', color=colors, edgecolor='none', ax=ax_bar, width=width)  # Set edgecolor to 'none' for no edges

    ax_bar.xaxis.set_visible(False)

    ax_bar.set_frame_on(False)
    plt.grid(False)

    # Hide y-axis label
    ax_bar.set_ylabel('')

    # Increase the size of the y-axis ticks
    ax_bar.tick_params(axis='y', which='major', labelsize=12)  # Adjust the labelsize as needed

    # Add names at the center or outside of each bar with a larger font size
    for bar, name in zip(bars.patches, value_counts.index):
        width, height = bar.get_width(), bar.get_height()
        count = int(width)
        proportion = width / total_count
        length_factor = (len(str(name))+2) * 0.045  # Adjust this factor to control the influence of name length

        
        if proportion / length_factor < threshold:
            ax_bar.text(width + 0.01, bar.get_y() + height/2, f' {name}', va='center', ha='left', color='#636363', size=14)
        else:
            ax_bar.text(width/2, bar.get_y() + height/2, f'{name}', va='center', ha='center', color=cfg['bar_font_color'], size=14)

    # Set a maximum of 6 ticks
    num_ticks = min(len(value_counts), 6)
    
    # Set y-axis ticks and labels
    ax_bar.set_yticks(range(num_ticks))
    # Assuming your keys are the same as the range of the length of the array
    ax_bar.set_yticklabels(value_counts.values)


    ax_bar.yaxis.set_visible(True)
    #print(value_counts.values)
    # Save figure to a buffer
    buffer_bar = BytesIO()
    fig_bar.savefig(buffer_bar, format='png')
    image_bar_thumb = base64.b64encode(buffer_bar.getvalue()).decode('utf-8')
    plt.close(fig_bar)

    return image_bar_thumb


def pie_chart(data_series, cfg, name=None):
    # Use value_counts to get counts and labels
    value_counts = data_series.value_counts()

    # Extract data and labels from the value_counts result
    data = value_counts.values
    labels = value_counts.index

    # Limit to the top 5 categories
    top_categories = 5
    if len(labels) > top_categories:
        data = data[:top_categories]
        labels = labels[:top_categories]
        # Combine the remaining categories into "Others"
        data = np.append(data, sum(value_counts.values[top_categories:]))
        labels = np.append(labels, 'Others')

    # Create a pie chart
    fig_pie, ax_pie = plt.subplots(figsize=(6, 4))
    ax_pie.pie(data, labels=labels, autopct='%1.1f%%', colors=sns.color_palette('pastel'), startangle=90)

    ax_pie.tick_params(axis='both', colors='#a9a9a9')
    if name is not None:
        ax_pie.set_title(name)
    ax_pie.set_frame_on(False)
    plt.grid(False)

    # Save the figure to a buffer
    buffer_pie = BytesIO()
    fig_pie.savefig(buffer_pie, format='png')
    image_pie_thumb = base64.b64encode(buffer_pie.getvalue()).decode('utf-8')
    plt.close(fig_pie)

    return image_pie_thumb

def Nightndale_chart(data_series, cfg, name=None):
    #Use value_counts to get counts and labels
    value_counts = data_series.value_counts()

    # Extract data and labels from the value_counts result
    data = value_counts.values
    labels = value_counts.index

    # Limit to the top 5 categories
    top_categories = 10
    if len(labels) > top_categories:
        data = data[:top_categories]
        labels = labels[:top_categories]
        # Combine the remaining categories into "Others"
        data = np.append(data, sum(value_counts.values[top_categories:]))
        labels = np.append(labels, 'Others')

    # Convert data to radians
    theta = np.linspace(0.0, 2 * np.pi, len(labels), endpoint=False)

    # Create a Nightingale diagram
    fig, ax = plt.subplots(subplot_kw={'projection': 'polar'}, figsize=(6, 4))
    ax.fill(theta, data, color='skyblue', alpha=0.75)

    # Customize the plot
    ax.set_theta_offset(np.pi / 2)  # Adjust the starting angle if needed
    ax.set_theta_direction(-1)  # Change to 1 for clockwise direction
    ax.set_rlabel_position(0)  # Move radial labels to the center
    ax.set_yticklabels([])  # Hide radial labels if not needed

    # Add category labels
    ax.set_xticks(theta)
    ax.set_xticklabels(labels)

    if name is not None:
        ax.set_title(name)

    # Save the figure to a buffer
    buffer_nightingale = BytesIO()
    fig.savefig(buffer_nightingale, format='png')
    image_nightingale_thumb = base64.b64encode(buffer_nightingale.getvalue()).decode('utf-8')
    plt.close(fig)

    return image_nightingale_thumb



def radial_bar_chart(data_series,cfg):
    # Extract up to 5 categories and values from the data series
    value_counts = data_series.value_counts()

    # Extract data and labels from the value_counts result
    data = value_counts.values
    labels = value_counts.index

    # Limit to the top categories
    top_categories = 10
    if len(labels) > top_categories:
        data = data[:top_categories]
        labels = labels[:top_categories]
        # Combine the remaining categories into "Others"
        data = np.append(data, sum(value_counts.values[top_categories:]))
        labels = np.append(labels, 'Others')

    theta = np.linspace(0.0, 2 * np.pi, len(labels), endpoint=False)

    fig, ax = plt.subplots(subplot_kw={'projection': 'polar'})
    ax.bar(theta, data, color=['skyblue', 'orange', 'lightgreen', 'pink', 'lightcoral'], alpha=0.7)

    ax.set_theta_zero_location('N')
    ax.set_theta_direction(-1)
    ax.set_rlabel_position(0)  # Move radial labels to the center
    ax.set_yticklabels([])  # Hide radial labels if not needed

    ax.set_xticks(theta)
    ax.set_xticklabels(labels)

    plt.title('Radial Bar Chart')

    # Save figure to a buffer
    buffer_radial = BytesIO()
    plt.savefig(buffer_radial, format='png')
    image_radial = base64.b64encode(buffer_radial.getvalue()).decode('utf-8')
    plt.close(fig)

    return image_radial




def treemap(data_series, cfg):
    # Use value_counts to get counts and labels
    value_counts = data_series.value_counts()

    # Extract data and labels from the value_counts result
    data = value_counts.values
    labels = value_counts.index
    top_categories=5
    # Limit to the top categories
    if len(labels) > top_categories:
        data = data[:top_categories]
        labels = labels[:top_categories]
        # Combine the remaining categories into "Others"
        data = np.append(data, sum(value_counts.values[top_categories:]))
        labels = np.append(labels, 'Others')

    squarify.plot(sizes=data, label=labels, color=['skyblue', 'orange', 'lightgreen', 'pink', 'lightcoral'], alpha=0.7)
    plt.axis('off')
    plt.title('Treemap')

    # Save figure to a buffer
    buffer_treemap = BytesIO()
    plt.savefig(buffer_treemap, format='png')
    image_treemap = base64.b64encode(buffer_treemap.getvalue()).decode('utf-8')
    plt.close()

    return image_treemap



def donut_chart(data_series, cfg):
    # Use value_counts to get counts and labels
    value_counts = data_series.value_counts()

    # Extract data and labels from the value_counts result
    data = value_counts.values
    labels = value_counts.index
    top_categories=5
    # Limit to the top categories
    if len(labels) > top_categories:
        data = data[:top_categories]
        labels = labels[:top_categories]
        # Combine the remaining categories into "Others"
        data = np.append(data, sum(value_counts.values[top_categories:]))
        labels = np.append(labels, 'Others')

    plt.pie(data, labels=labels, autopct='%1.1f%%', startangle=90, wedgeprops=dict(width=0.4), colors=['skyblue', 'orange', 'lightgreen', 'pink', 'lightcoral'])
    plt.axis('equal')  # Equal aspect ratio ensures the donut is circular.
    plt.title('Donut Chart')

    # Save figure to a buffer
    buffer_donut = BytesIO()
    plt.savefig(buffer_donut, format='png')
    image_donut = base64.b64encode(buffer_donut.getvalue()).decode('utf-8')
    plt.close()

    return image_donut

##################################################################################################################
def hist_plot_thumb(data,cfg, name=None):
    # Create a histogram
    fig_histogram, ax_histogram = plt.subplots(figsize=(6, 4))
    ax_histogram.hist(data, bins=30, color=cfg['hist_color'], edgecolor='white', density=True) #'skyblue'
    
    # Add a kernel density plot
    sns.kdeplot(data, color=cfg['hist_kde_color'], ax=ax_histogram, warn_singular=False) #'pink'
    
    ax_histogram.tick_params(axis='both', colors='#a9a9a9')
    ax_histogram.set_xlabel('Value', color='#a9a9a9')
    ax_histogram.set_ylabel('Frequency', color='#a9a9a9')
    if name is not None:
        ax_histogram.set_title(name)
    ax_histogram.set_frame_on(False)
    plt.grid(False)

    # Save the figure to a buffer
    buffer_histogram = BytesIO()
    fig_histogram.savefig(buffer_histogram, format='png')
    image_histogram_thumb = base64.b64encode(buffer_histogram.getvalue()).decode('utf-8')
    plt.close(fig_histogram)

    return image_histogram_thumb


###########################################################################################################################
######################################### WRAPPER ###########################################################
###########################################################################################################################




def missing_plot(data, cfg, name=None):
    # Calculate the percentage of missing values for each column
    missing_percentage = data.isnull().mean() * 100
    
    # Create a bar plot with missing values on the upper part and rotated bottom ticks
    fig, ax = plt.subplots(figsize=(12, 4))  # Adjusted the width to provide more space
    
    # Set the bar width (adjust this value as needed)
    bar_width = 0.5
    
    # Plot blue bars for non-missing values
    ax.bar(missing_percentage.index, 100 - missing_percentage, color=cfg['hist_color'], label='All Values', width=bar_width)
    
    # Plot red bars for missing values on the upper part
    ax.bar(missing_percentage.index, missing_percentage, bottom=100 - missing_percentage, color=cfg['hist_kde_color'], label='Missing Values', width=bar_width)
    
    # Customize the plot
    ax.set_ylabel('Percentage')
    ax.set_title('Missing Values Bar Chart',pad=20)
    ax.legend()
    
    # Rotate the bottom ticks for better readability
    plt.xticks(rotation=45, ha='right') 
    
    # Adjust layout to prevent x-axis tick names from being cut out
    plt.tight_layout()
    
    # Remove the frame and grid
    ax.set_frame_on(False)
    plt.grid(False)

    # Save the figure to a buffer
    buffer_plot = BytesIO()
    fig.savefig(buffer_plot, format='png')
    missing_plot = base64.b64encode(buffer_plot.getvalue()).decode('utf-8')
    plt.close(fig)

    return missing_plot


###########################################################################################################################
######################################### DASH ###########################################################
###########################################################################################################################


def triple_donut(data, cfg, name=None):
    #print(data)
    # Data for the three donut charts
    data1 = data['Missing cells (%)']  # Replace with your desired percentage
    data2 = data['Duplicate rows (%)']  # Replace with your desired percentage
    data3 = data['Constant columns (%)'] # Replace with your desired percentage
    
    # Colors for the donut charts
    color1 = '#DCDCDC' # Background color
    color2 = cfg['dash_donuts_color']      # Color corresponding to the value
    
    # Create subplots with three donut charts
    fig, ax = plt.subplots(1, 3, figsize=(6, 1.75))
    
    # Function to draw a donut chart with a percentage value in the center
    def draw_donut(ax, data, title):
        ax.pie([100 - data, data], colors=[color1, color2], autopct='', startangle=90,
               wedgeprops=dict(width=0.3, edgecolor='w'))  # Add wedgeprops to create a hole in the center
    
        # Add the percentage value in the center
        ax.text(0, 0, f'{data}%', ha='center', va='center', fontsize=16, weight='bold', color=color2)
    
        ax.set_title(title)
    
    # Donut chart 1
    draw_donut(ax[0], data1, 'Missing cells')
    
    # Donut chart 2
    draw_donut(ax[1], data2, 'Duplicate rows')
    
    # Donut chart 3
    draw_donut(ax[2], data3, 'Constant columns')
    
    # Equal aspect ratio ensures that pie is drawn as a circle.
    for a in ax:
        a.axis('equal')
    
    # Save the figure to a buffer
    buffer_image = BytesIO()
    fig.savefig(buffer_image, format='png')
    image_thumb = base64.b64encode(buffer_image.getvalue()).decode('utf-8')
    plt.close(fig)


    return image_thumb



def triple_box(data, cfg, name=None):
    # Create a histogram
    fig, ax = plt.subplots(figsize=(6, 2))

    values = {'Categorical':data['Categorical'],
            'Numeric':data['Numeric'],
            'Text':data['Text'],
            'Datetime':data['Datetime'],
    }
    # Assuming 'data' is a list of four values

    # Filter out values that are <= 0
    filtered_values = [value for k,value in values.items() if value > 0]

    # Custom bar labels (you can modify these as needed)
    bar_labels = [k for k,value in values.items() if value > 0]
    
    # Define a color palette for bars
    #colors = sns.color_palette('YlOrBr', len(filtered_values))
    col_in = "light:" + cfg['dash_bars_color']
    colors = sns.color_palette(col_in,len(filtered_values))
    
    bars = ax.barh(bar_labels, filtered_values, color=colors, edgecolor='white', alpha=0.7)

    # Add labels inside the bars for values > 0
    for bar, value, label in zip(bars, filtered_values, bar_labels):
        if value > 0:
            text = f'{label}: {value}'
            ax.text(value / 2, bar.get_y() + bar.get_height() / 2, text, va='center', ha='center', color=cfg['dash_bars_text_color'])

    ax.tick_params(axis='both', colors='#a9a9a9')
    ax.set_xlabel('Values', color='#a9a9a9')
    ax.set_ylabel('Category', color='#a9a9a9')

    # Set y-axis ticks and labels
    ax.set_yticks(range(len(filtered_values)))
    ax.set_yticklabels(filtered_values)

    if name is not None:
        ax.set_title(name)

    ax.set_frame_on(False)
    plt.grid(False)

    # Save the figure to a buffer
    buffer_img = BytesIO()
    fig.savefig(buffer_img, format='png')
    image_thumb = base64.b64encode(buffer_img.getvalue()).decode('utf-8')
    plt.close(fig)

    return image_thumb




    
#data1 = { 'Categorical': data['Categorical'],
            # 'Numeric': data['Numeric'],
            # 'Text': data['Text'],
            # 'Datetime': data['Datetime'],
            #  }

    

# Number of variables 12
# Number of observations 891
# Missing cells 866
# Missing cells (%) 0.97
# Duplicate rows 0
# Duplicate rows (%) 0.0
# Total size in memory 85668
# Avg. record size in memory 96.14814814814815
# Numeric 3
# Categorical 6
# Text 3
# Datetime 0
# missing {'Age': 177, 'Cabin': 687, 'Embarked': 2}
# unique {'PassengerId': 891, 'Name': 891}
# constant {}
# zero {'Survived': 549, 'SibSp': 608, 'Parch': 678, 'Fare': 15}
# negative {}


        
def wiz_wrap(hub,cfg):
    """
    Create visualizations for plot(all_df)
    """
    
    dfw, data = hub["col"], hub["data"]

    # im passing the v_type from comp already, here inseted of continuous, it changes is to numerical
    #v_type = 'numerical'
    #data['v_type'] = v_type
    
    #pic_thumb = hist_plot_thumb(col,cfg)
    pic_thumb = missing_plot(dfw,cfg)
    #toto = tab_info(hub,cfg)

    donut = triple_donut(data,cfg)
    dash_b = triple_box(data,cfg)
    #print('df')

    
    
    # Store the results in variables to avoid repeated dictionary lookups
    stat_tab = generate_table(data, container_keysW, 'val_tableW',style, pic_thumb, cfg=cfg)
    val_tabx = value_tableW(dfw,cfg)
    #qq = qq_plot(col,cfg)

    plots = {
        'type': ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm'],
        'title': ['Dash','Stats', 'Missing', 'Value Table'],
        'image': [((donut,500),(dash_b,500),),None, ((pic_thumb,1000),), None],
        'image_width': 1000,
        'value_table': [None,stat_tab, None, val_tabx,None],
        
    }
    
    return plots

def wiz_biv_cat(hub):
    """
    Create visualizations for plot(df, Continuous)
    """

    col1,col2, data = hub["col1"], hub["col2"], hub["data"]
    #print(col,data)
    #tabs: List[Panel] = []
    # pic = hist_plot(col)
    # pic2 = qq_plot(col)
    # pic3 = kde_plot(col)
    #htgs: Dict[str, List[Tuple[str, str]]] = {}

    # Set the width and height for both charts
    width, height = 8, 6



#aaa


    store = {
        # 'val_tab2': value_tableH(data,pic),
        # 'val_tab':value_tableH(data,pic2,pic3),
        'hm': heatmap_plot(col1,col2),
        'nbch' : nested_bar_chart(col1,col2,width, height),
        'sbch'  : stacked_bar_chart_percentage(col1,col2,width, height),
        # 'box' : box_plot(col),
    }

    plots = {
            'type': ['a','b','c','d','e','f','g','h','i','j','k','l','m'
            ],
        
            'title': [#'iq',
                      #'Value Table', 
                      'HM', 
                      'Nest', 
                      'Stacked', 
                      #'Box Plot'
                    ],
        
            'image': [#None, 
                      #None, 
                      ((store['hm'],800),), #800 for all
                      ((store['nbch'],800),), 
                      ((store['sbch'],800),),
                      #store['box']
                    ],
        
            'value_table': [#store['val_tab2'],
                            #store['val_tab'], 
                            None, 
                            None, 
                            None, 
                            #None
            ],
        }
    
    return plots
def wiz_biv_con_cat(hub):
    """
    Create visualizations for plot(df, Continuous)
    """
    col1,col2, data = hub["col1"], hub["col2"], hub["data"]
    #print(col1.name)

    #(('categorical', 2, 0.002244668911335578), ('continuous', 88, 0.09876543209876543))
    if data[0][0] == 'continuous':
        col1,col2 = col2,col1
    #print(data)
    
    #print(col,data)
    #tabs: List[Panel] = []
    # pic = hist_plot(col)
    # pic2 = qq_plot(col)
    # pic3 = kde_plot(col)
    #htgs: Dict[str, List[Tuple[str, str]]] = {}






    store = {
        # 'val_tab2': value_tableH(data,pic),
        # 'val_tab':value_tableH(data,pic2,pic3),
        'vp'  : violin_plot(col1,col2),
        'bx': box_plot_con(col1,col2),
        #'lc' : line_chart(col1,col2),
        # 'box' : box_plot(col),
    }

    plots = {
            'type': ['a','b','c','d','e','f','g','h','i','j','k','l','m'
            ],
        
            'title': [#'iq',
                      #'Value Table', 
                      'Violin', 
                      'Box', 
                      #'Line', 
                      #'Box Plot'
                    ],
        
            'image': [#None, 
                      #None, 
                      ((store['vp'],800),),
                      ((store['bx'],800),), 
                      #((store['lc'],800),), 
                      #store['box']
                    ],
        
            'value_table': [#store['val_tab2'],
                            #store['val_tab'], 
                            None, 
                            None, 
                            #None, 
                            #None
            ],
        }
    
    return plots
#axxx
def wiz_biv_con(hub):
    """
    Create visualizations for plot(df, Continuous)
    """
    col1,col2, data = hub["col1"], hub["col2"], hub["data"]
    #print(col,data)
    #tabs: List[Panel] = []
    # pic = hist_plot(col)
    # pic2 = qq_plot(col)
    # pic3 = kde_plot(col)
    #htgs: Dict[str, List[Tuple[str, str]]] = {}






    store = {
        # 'val_tab2': value_tableH(data,pic),
        # 'val_tab':value_tableH(data,pic2,pic3),
        'sc': scatter_plot(col1,col2),
        'hx' : hexbin_plot(col1,col2),
        # 'qq'  : qq_plot(col),
        # 'box' : box_plot(col),
    }

    plots = {
            'type': ['a','b','c','d','e','f','g','h','i','j','k','l','m' 
                     
            ],
        
            'title': [#'iq',
                      #'Value Table', 
                      'Scatter', 
                      'Hexbin', 
                      #'QQ Plot', 
                      #'Box Plot'
                    ],
        
            'image': [#None, 
                      #None, 
                      ((store['sc'],800),), 
                      ((store['hx'],800),), 
                      #store['qq'],
                      #store['box']
                    ],
        
            'value_table': [#store['val_tab2'],
                            #store['val_tab'], 
                            None, 
                            None, 
                            #None, 
                            #None
            ],
        }
    
    return plots

container_keysN = {
        'container1': ['VALUES', 'MISSING', 'DISTINCT', '', 'MEMORY', 'DTYPE'],
        'container2': ['MAX', '95%', 'Q3', 'AVG', 'MEDIAN', 'Q1', '5%', 'MIN'],
        'container3': ['RANGE', 'IQR', 'STD', 'VAR', ' ', 'KURT.', 'SKEW', 'SUM']
    }

container_keysT = {
        'container1': ['VALUES', 'MISSING', 'DISTINCT', '', 'MEMORY', 'DTYPE'],
        'container2': ['Max length', 'Mean length','Median length', 'Min length'],
        'container3': ['RANGE', 'IQR', 'STD', 'VAR', ' ', 'KURT.', 'SKEW', 'SUM']
    }

container_keysC = {
        'container1': ['VALUES', 'MISSING', 'DISTINCT', '', 'MEMORY', 'DTYPE'],
        'container3a': [0,1,2,3,4,5]
        #'container3': ['RANGE', 'IQR', 'STD', 'VAR', ' ', 'KURT.', 'SKEW', 'SUM']
    }

container_keysW = {
        'Stats': ['Number of columns', 'Number of rows', 'Missing cells', 'Missing cells (%)', 'Duplicate rows', 'Duplicate rows (%)','Total memory size in KB','Avg. record size in Bytes'],
        'Columns': ['Numeric', 'Categorical', 'Text', 'Datetime',],
        'Alerts': ['missing', 'unique', 'constant', 'zero', 'negative'],
    }



def wiz_text(hub,cfg):

    """
    Create visualizations for plot(df, Text)
    """

    col, data = hub["col"], hub["data"]

    w_len = data['w_len']
    #pic = hist_plot_thumb(w_len,cfg,name="Value Lengths")
    
    pic, tokens = word_cloud(col, cfg)

    


    from collections import Counter
    
    # Basic Analysis: Calculate word frequencies
    #word_freq = Counter(tokens)
    #most_common_words = word_freq.most_common(10)
    #print("Most common words:", most_common_words)
    serial_tokens = pd.Series(tokens).sort_values(ascending=False)

    
    # print(len(tokens))
    # print(serial_tokens.value_counts())
    # print(col)
    #print(ser_tokens)

    
    store = {
        'stat_tab': generate_table(data,container_keysT,'val_tableT', style, pic,cfg=cfg),
        #'val_tab1': value_tableT(data,style,pic),
        'val_tabx':value_tableX(col,cfg,tokens=serial_tokens),
        
        #'hist': hist_plot(col),
        #'kde' : kde_plot(col),
        #'qq'  : qq_plot(col),
        #'box' : box_plot(col),
    }
    toto = tab_info(hub,cfg)

    plot_info = {
        'type': ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm'],
        'title': [toto, 'Stats', 'Value Table'],
        'image': [None, None, None],
        'value_table': [None, store['stat_tab'], store['val_tabx']],
    }

    return plot_info


def wiz_cat(hub,cfg):
    """
    Create visualizations for plot(df, Categorical)
    """
    col, data = hub["col"], hub["data"]
    
    # Pass the v_type from comp already; here instead of categorical, it changes to categorical
    # v_type = 'categorical'
    # data['v_type'] = v_type

    pic = bar_chart_thumb(col,cfg)
    
    pie = pie_chart(col,cfg)
    #rad = radial_bar_chart(col,cfg)

    fake_dict = {1:'a',2:'b',3:'c',4:'d',5:'e'}
    data.update(fake_dict)
    
    toto = tab_info(hub,cfg)
    
    # Store the results in variables to avoid repeated dictionary lookups
    stat_tab = generate_table(data, container_keysC, 'val_tableC', style, pic,cfg=cfg)
    val_tabx = value_tableX(col,cfg)
    #hist = hist_plot(col,cfg)

    plots = {
        'type': ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm'],
        'title': [toto, 'Stats', 'Pie Chart', 'Value Table'],
        'image': [None, None, ((pie,500),), None],
        'value_table': [None, stat_tab, None, val_tabx],
    }
    
    return plots

def wiz_num(hub,cfg):
    """
    Create visualizations for plot(df, Continuous)
    """
    
    col, data = hub["col"], hub["data"]

    # im passing the v_type from comp already, here inseted of continuous, it changes is to numerical
    #v_type = 'numerical'
    #data['v_type'] = v_type
    
    pic_thumb = hist_plot_thumb(col,cfg)
    
    toto = tab_info(hub,cfg)
    
    # Store the results in variables to avoid repeated dictionary lookups
    stat_tab = generate_table(data, container_keysN, 'val_tableH', style, pic_thumb,cfg=cfg)
    val_tabx = value_tableX(col,cfg)
    qq = qq_plot(col,cfg)

    plots = {
        'type': ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm'],
        'title': [toto, 'Stats', 'QQ Plot', 'Value Table'],
        'image': [None, None, ((qq,500),), None],
        'value_table': [None, stat_tab, None, val_tabx],
    }
    
    return plots

def wiz_dt(hub,cfg):

    """
    Create visualizations for plot(df, dt)
    """
    col, data = hub["col"], hub["data"]
    
    # Pass the v_type from comp already; here instead of categorical, it changes to categorical
    # v_type = 'categorical'
    # data['v_type'] = v_type

    pic = hist_plot_thumb(col,cfg)
    #pic = bar_chart_thumb(col,cfg)
    #pie = pie_chart(col,cfg)
    
    toto = tab_info(hub,cfg)
    
    # Store the results in variables to avoid repeated dictionary lookups
    stat_tab = generate_table(data, container_keysN, 'val_tableH', style, pic,cfg=cfg)
    val_tabx = value_tableX(col,cfg)
    #hist = hist_plot(col,cfg)

    plots = {
        'type': ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm'],
        'title': [toto, 'Stats', 'Value Table'],
        'image': [None, None, None],
        'value_table': [None, stat_tab, val_tabx],
    }
    
    return plots

def tab_info(hub,cfg):

    #print(hub)
    name = hub["col"].name    
    col, data = hub["col"], hub["data"]
    

    """" just copy of render num """
    pic_mini = hist_plot_thumb_mini(col,cfg)
    
    if data['MISSING_P'] > 0:
        font_size = '13px'
        font_color = '#FF6961'
        font_weight = '500'
    else:
        font_size = '11px'
        font_color = '#a9a9a9'
        font_weight = '300'
        
    if data['DISTINCT'] == 1 or data['DISTINCT_P'] == 100:
        font_sizeD = '13px'
        font_colorD = '#FF6961' #red
        font_weightD = '500'
    else:
        font_sizeD = '11px'
        font_colorD = '#a9a9a9'
        font_weightD = '300' 


    return f"""
        <div class="top_tab" style="width: 750px; height: 22px; padding: 0px 10px 0px; background-color: white; display: flex; align-items: center;">
            <div style="padding: 12px 10px 0px; width: 270px;">
                    <div style="font-size: 18px; height:20px;">{name}</div>
                    <div class="bottom_line" style="font-size: 12px; color: #a9a9a9; font-weight: 500; height:20px;">{data['v_type']}</div>
            </div>
                           <div class="table_tab" style="width: 180px; font-size: 10px; margin-bottom: 0px;">
                                <div style="display: flex;">
                        
                                    <div style="width: 60px; font-size: 11px; text-align: center; padding: 10px 5px; color: #a9a9a9; font-weight: 300;"> {data['VALUES_P']}%</div>
                                     <div style="width: 60px; font-size: {font_size}; text-align: center; padding: 10px 5px; color: {font_color}; font-weight: {font_weight};"> {data['MISSING_P']}%</div>
                                    <div style="width: 60px; font-size: {font_sizeD}; text-align: center; padding: 10px 5px; color: {font_colorD}; font-weight: {font_weightD};"> {data['DISTINCT_P']}%</div>
                                   
                                </div>
                                <div style="display: flex;">
                        
                                    
                                    <div style="width: 60px; font-size: 10px; text-align: center; padding: 0px 5px; color: #a9a9a9; font-weight: 300;"> VALID</div>
                                    <div style="width: 60px; font-size: 10px; text-align: center; padding: 0px 5px; color: #a9a9a9; font-weight: 300;"> MISSING</div>
                                    <div style="width: 60px; font-size: 10px; text-align: center; padding: 0px 5px; color: #a9a9a9; font-weight: 300;"> UNIQUE</div>
                                   
                                </div>
                            </div>
                       
            <div style="padding: 0px 10px 0px; width: 300px;">
                        <div class="table_tab_pic" style="width: 300px; height: 50px; font-size: 10px; margin-bottom: 0px;">
                            <div style="text-align: center; padding: 0px 5px; color: #a9a9a9; font-weight: 300; height: 40px;">
                                <img class="img-thumb-mini" src="data:image/png;base64, {pic_mini}" alt="Histogram" style="padding-top: 5px;">
                            </div>
                                <div style="display: flex;">
                                    <div style="width: 50px;"></div>
                                    <div style="width: 50px; font-size: 10px; text-align: left; padding: 0px 5px; color: #a9a9a9; font-weight: 300;"> Unique</div>
                                    <div style="width: 100px; font-size: 11px; text-align: left; padding: 0px 5px; color: {font_colorD}; font-weight: 500;"> {data['DISTINCT']}</div>
                                    <div style="width: 50px;"></div>
                                    <div style="width: 50px; font-size: 10px; text-align: left; padding: 0px 5px; color: #a9a9a9; font-weight: 300;"> Total</div>
                                    <div style="width: 100px; font-size: 11px; text-align: left; padding: 0px 5px; color: #a9a9a9; font-weight: 500;"> {data['TOTAL']}</div>
                                </div>
        
                        </div>
            </div>
        </div>
            """

def wiz(hub,cfg):
    """
    Render a basic plot
    Parameters
    ----------
    hub
        The DataHub containing results from the compute function.
    cfg
        Config instance
    """
    if hub.variable_type == "categorical":
        variable_data = wiz_cat(hub,cfg)
    elif hub.variable_type == "continuous":
        variable_data = wiz_num(hub,cfg)
    elif hub.variable_type == "text":
        variable_data = wiz_text(hub,cfg)
    elif hub.variable_type == "biv_con":
        variable_data = wiz_biv_con(hub)
    elif hub.variable_type == "biv_con_cat":
        variable_data = wiz_biv_con_cat(hub)
    elif hub.variable_type == "biv_cat":
        variable_data = wiz_biv_cat(hub)
    elif hub.variable_type == "datetime":
        variable_data = wiz_dt(hub, cfg)
    elif hub.variable_type == "wrapper":        
        variable_data = wiz_wrap(hub, cfg)
        
    # elif hub.variable_type == "geography_column":
    #     variable_data = render_geo(hub, cfg)
    return variable_data


import os
from os.path import dirname
from typing import List

import pandas as pd

def get_dataset_names() -> List[str]:
    """
    Get all available dataset names. It is all csv file names in 'data' folder.

    Returns
    -------
    datasets: list
        A list of all available dataset names.

    """
    module_path = dirname(__file__)
    files = os.listdir(module_path)
    csv_files = [f for f in files if f.endswith(".csv")]

    # remove suffix csv and get dataset names
    datasets = [os.path.splitext(f)[0] for f in csv_files]

    return datasets


def _get_dataset_path(name: str) -> str:
    """
    Given a dataset name, output the file path.
    """
    # Remove suffix 'csv' and transform to lower case
    lower_name = name.lower()
    if lower_name.endswith(".csv"):
        lower_name = os.path.splitext(lower_name)[0]

    if lower_name not in get_dataset_names():
        raise ValueError(
            f"Dataset {name} is not found. You may want to try get_dataset_names()"
            + " to get all available dataset names"
        )

    module_path = dirname(__file__)
    path = os.path.join(module_path,f"{lower_name}.csv")
    return path


def load_dataset(name: str) -> pd.DataFrame:
    """
    Load dataset of the given name.

    Parameters
    ----------
    name: str
        Dataset name. The dataset will be loaded from 'data/{name}.csv'.

    Returns
    -------
    df : dataframe
        A dataframe of corresponding dataset.

    Examples
    --------
    Load titanic dataset:
    >>> from dataprep.datasets import load_dataset
    >>> df = load_dataset('titanic')

    Get all available dataset names:
    >>> from dataprep.datasets import get_dataset_names
    >>> get_dataset_names()
    ['iris', 'titanic', 'adult', 'house_prices_train', 'house_prices_test']
    """
    path = _get_dataset_path(name)
    df = pd.read_csv(path)
    return df
